# -*- coding: UTF-8 -*-

import grp
import os
import pymysql
import pwd
import re
import shutil
from nes_common import logger, rest_client
from nes_common.config import config
from nes_common.executor import Executor, DataIntegrityError


class TableManager(object):
    def __init__(self, cursor, log):
        self.log = log
        self.cursor = cursor
        self.workTables = {}
        self.sysUser = config.getString('mysql', 'sysUser')
        self.sysGroup = config.getString('mysql', 'sysGroup')
        self.tablesWithoutCheck = [table.strip() for table in config.getString('MAIN', 'tablesWithoutCheck').split(',')]
        self.path = self.getDBPath(cursor)
        self.log.debug('Init TableManager')

    def getDBPath(self, cursor):
        cursor.execute("SHOW VARIABLES WHERE Variable_Name = 'datadir'")
        row = cursor.fetchone()
        if not row:
            self.log.error('Не удалось получить путь до базы данных dictionaries')
            return ''
        return row[1] + 'dictionaries'

    def invertName(self, base, name):
        return base + '1' if name.endswith('2') else base + '2'

    def prepareWorkingCopy(self, tableBase, tableInUse):
        tableWorking = self.invertName(tableBase, tableInUse)

        self.cursor.execute("SELECT engine FROM information_schema.tables WHERE table_schema='dictionaries' "
                "AND table_name='%s'" % tableInUse)
        res = self.cursor.fetchone()[0]

        if res.lower() == "innodb":
            self.prepareWorkingCopyInnoDB(tableInUse, tableWorking)
        else:
            self.prepareWorkingCopyMyISAM(tableInUse, tableWorking)

        return tableWorking

    def prepareWorkingCopyMyISAM(self, tableInUse, tableWorking):
        try:
            self.log.info("dropping %s" % tableWorking)
            self.cursor.execute('DROP TABLE %s' % tableWorking)
        except Exception as e:
            self.log.error("can't drop %s" % tableWorking)
            self.log.error(e)

        fromPath = self.path + '/' + tableInUse
        toPath = self.path + '/' + tableWorking
        try:
            self.log.info("copying %s to %s" % (fromPath, toPath))
            for ext in [".frm", ".MYD", ".MYI"]:
                shutil.copyfile(fromPath + ext, toPath + ext)
                shutil.copymode(fromPath + ext, toPath + ext)
                try:
                    os.chown(toPath + ext, pwd.getpwnam(self.sysUser)[2], grp.getgrnam(self.sysGroup)[2])
                except:
                    pass
        except Exception as e:
            self.log.error("can't copy from %s to %s" % (fromPath, toPath))
            raise DataIntegrityError()

    def prepareWorkingCopyInnoDB(self, tableInUse, tableWorking):
        try:
            self.log.info('truncate table %s' % tableWorking)
            self.cursor.execute('TRUNCATE TABLE %s' % tableWorking)

            self.log.info('insert into %s select * from %s' % (tableWorking, tableInUse))
            self.cursor.execute('INSERT INTO %s SELECT * FROM %s' % (tableWorking, tableInUse))
        except Exception as e:
            self.log.error(e)

    def getWorkingCopy(self, table):
        if self.workTables.get(table, None) is None:
            try:
                self.cursor.execute('show create view %s' % (table,))
                row = self.cursor.fetchone()
                matches = re.search('%s[12]' % (table,), row[1])
                if matches is not None:
                    result = self.prepareWorkingCopy(table, matches.group(0))
                    self.log.info('will work with %s' % (result,))
                    self.workTables[table] = result
            except Exception as e:
                self.log.info('%s is not view' % (table,))
                self.log.exception(e)
                self.workTables[table] = table
        return self.workTables[table]

    def finalize(self):
        allOk = True
        self.log.info('checking tables')
        for view, table in self.workTables.items():
            if view not in self.tablesWithoutCheck:
                self.log.info('check table %s' % table)
                self.cursor.execute('check table %s' % table)
                for row in self.cursor:
                    if row[2] == 'status':
                        self.log.info('%s status is %s' % (table, row[3]))
                        allOk = (row[3] == 'OK') and allOk
        if allOk:
            self.log.info('all tables are OK. altering views')
            for view, table in self.workTables.items():
                if view != table:
                    self.cursor.execute('CREATE OR REPLACE VIEW %s AS SELECT * FROM %s' % (view, table))
            self.log.info('views are altered now')
        return allOk


class Context(object):
    """ Контекст выполнения SQL запросов"""

    def __init__(self, cursor, log):
        """ Контекст выполнения SQL запросов
        Args:
            cursor (Cursor): Соединения с базой данных
        """
        self.cursor = cursor
        self.log = log
        self.tableManager = TableManager(self.cursor, self.log)
        self.emptyTables = {}
        self.log.debug('Init context')

    def tableName(self, table):
        """ Получить окончательное название таблицы
        Args:
            table (str): Базовое название таблицы
        Returns:
            str: Окончательное название таблицы
        """
        return self.tableManager.getWorkingCopy(table)

    def finalize(self):
        """ Выполнить завершающие операции
        Returns:
            bool: Успешно или нет
        """
        return self.tableManager.finalize()


class Exec(Executor):
    def __init__(self, host, port, user, password, database, shopCode, loadDictStatusSender=None):
        self.log = logger.getLogger('Shop', shopCode)
        self.log.info('Connect to %s:%s %s' % (host, port, database))

        if not password or password == '':
            connection = pymysql.connect(host=host, port=port, user=user, db=database, charset='utf8')
        else:
            connection = pymysql.connect(host=host, port=port, user=user, password=password, db=database, charset='utf8')

        context = Context(connection.cursor(), self.log)
        self.log.info('Init Executor')
        loadDictStatusSender.sendDictState('Start load dict to db', 'Загрузка справочника в БД кассы', rest_client.STATUS_DURING)
        Executor.__init__(self, shopCode, connection, context)
